import torch 

def block(ci, co, k=5, s=1, p=0):
    return torch.nn.Sequential(
        torch.nn.Conv2d(ci, co, k, s, p),
        torch.nn.Tanh(),
        torch.nn.AvgPool2d(2, stride=2)
    )

class LeNet5(torch.nn.Module):
  def __init__(self, in_channels=1, n_classes=10):
    super().__init__()
    self.conv1 = block(in_channels, 6)
    self.conv2 = block(6, 16)
    self.conv3 = torch.nn.Sequential(
        torch.nn.Conv2d(16, 120, 5, padding=0),
        torch.nn.Tanh()
    )
    self.fc1 = torch.nn.Sequential(
        torch.nn.Linear(120, 84),
        torch.nn.Tanh()
    )
    self.fc2 = torch.nn.Linear(84, n_classes)

  def forward(self, x):
    x = self.conv1(x)
    x = self.conv2(x)
    x = self.conv3(x)
    x = x.view(x.shape[0], -1)
    x = self.fc1(x)
    x = self.fc2(x)
    return x


def accuracy(preds, targets):
	return (preds.argmax(dim=1) == targets).float().mean()

def error(preds, targets):
	return 1. - accuracy(preds, targets)

# top k error
# https://discuss.pytorch.org/t/top-k-error-calculation/48815/2
def topk_error(preds, targets, k):
	_, y_pred = preds.topk(k, dim=1)
	y_pred = y_pred.t()
	target_reshaped = targets.view(1, -1).expand_as(y_pred)
	correct = (y_pred == target_reshaped)
	ind_which_topk_matched_truth = correct[:k]
	flattened_indicator_which_topk_matched_truth = ind_which_topk_matched_truth.reshape(-1).float()
	tot_correct_topk = flattened_indicator_which_topk_matched_truth.float().sum(dim=0, keepdim=True)
	return 1. - tot_correct_topk / targets.shape[0]

def top5_error(preds, targets):
	return topk_error(preds, targets, 5)


import torch

class BarlowTwins(torch.nn.Module):

    def __init__(self, backbone, f1=512, f2=2048):
        super().__init__()
        self.backbone = backbone
        self.head = torch.nn.Sequential(
            torch.nn.Linear(f1, f2),
            torch.nn.BatchNorm1d(f2),
            torch.nn.ReLU(),
            torch.nn.Linear(f2, f2),
            torch.nn.BatchNorm1d(f2),
            torch.nn.ReLU(),
            torch.nn.Linear(f2, f2)
        )

    def forward(self, x):
        x = self.backbone(x)
        x = x.view(x.size(0), -1)
        x = self.head(x)
        return x


from pathlib import Path
import numpy as np
from .utils import download_data, generate_classes_list, generate_df
from sklearn.model_selection import train_test_split
import pandas as pd

class EuroSAT():
    def __init__(self, path="./data", val_size=0, test_size=0.2, random_seed=42, verbose=True, label_ratio=1):
        super().__init__()
        self.path = Path(path)
        self.num_classes = 10
        self.val_size = val_size
        self.test_size = test_size
        self.random_seed = random_seed
        self.url = "http://madm.dfki.de/files/sentinel/EuroSAT.zip"
        self.compressed_data_filename = 'EuroSAT.zip'
        self.data_folder = '2750'
        self.in_chans = 3
        self.verbose = verbose
        self.label_ratio = label_ratio
        # download and uncompress data
        uncompressed_data_path = download_data(
            self.path,
            self.compressed_data_filename,
            self.data_folder,
            self.url,
            self.verbose
        )
        self.classes = generate_classes_list(uncompressed_data_path)
        assert len(self.classes) == self.num_classes
        self.data = generate_df(self.classes, uncompressed_data_path, self.verbose)
        # make splits
        if self.test_size > 0:
            train_df, self.test = train_test_split(
                self.data,
                test_size=int(len(self.data)*self.test_size),
                random_state=self.random_seed,
                stratify=self.data.label.values
            )
        else: 
            train_df, self.test = self.data, None
        if self.val_size > 0:
            self.train, self.val = train_test_split(
                train_df,
                test_size=int(len(self.data)*self.val_size),
                random_state=self.random_seed,
                stratify=train_df.label.values
            )
        else: 
            self.train, self.val = train_df, None
        if self.verbose:
            print("Training samples", len(self.train))
            if self.val is not None:
                print("Validation samples", len(self.val))
            if self.test is not None:
                print("Test samples", len(self.test))
        # filter by label ratio (used for ssl validation experiments)
        if self.label_ratio < 1:
            train_labels = self.train.label.values
            train_images = self.train.image.values
            train_images_ratio, train_labels_ratio = [], []
            unique_labels = np.unique(train_labels)
            for label in unique_labels:
                filter = np.array(train_labels) == label
                ixs = filter.nonzero()[0]
                num_samples = filter.sum()
                ratio_ixs = np.random.choice(
                    ixs, int(self.label_ratio*num_samples), replace=False)
                train_images_ratio += (np.array(train_images)
                                       [ratio_ixs]).tolist()
                train_labels_ratio += (np.array(train_labels)
                                       [ratio_ixs]).tolist()
            self.train = pd.DataFrame(
                {'image': train_images_ratio, 'label': train_labels_ratio})
            if self.verbose:
                print("training samples after label ratio filtering:",
                      len(self.train))


import torch.nn as nn 
import torch 
import torchvision
from einops import rearrange

# only works for 224x224x3 inputs !

class Alexnet(nn.Module):
    def __init__(self, config=None, **kwargs):
        super(Alexnet, self).__init__()
        use_bn = config['batch_norm'] if config is not None and 'batch_norm' in config else False
        self.backbone = nn.Sequential(
            nn.Conv2d(3, 96, kernel_size=11, stride=4, padding=2),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(96) if use_bn else nn.Identity(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(96, 256, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(256) if use_bn else nn.Identity(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(256, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(384) if use_bn else nn.Identity(),
            nn.Conv2d(384, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(384) if use_bn else nn.Identity(),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(256) if use_bn else nn.Identity(),
            nn.MaxPool2d(kernel_size=3, stride=2),
        ) 
        self.head = nn.Sequential(
            nn.Flatten(),
            nn.Linear(256 * 6 * 6, 4096),
            nn.ReLU(inplace=True),
            nn.BatchNorm1d(4096) if use_bn else nn.Identity(),
            nn.Dropout(p=0.5),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.BatchNorm1d(4096) if use_bn else nn.Identity(),
            nn.Dropout(p=0.5),
            nn.Linear(4096, 1000),
        )

    def forward(self, x):
        return self.head(self.backbone(x))
    
    def tta(self, x):
        B = x.shape[0]                                                      # B, C, H, W    
        crops = torch.stack(torchvision.transforms.FiveCrop(224)(x))        # 5, B, C, H, W
        flips = torch.stack([torch.flip(crop, (-1,)) for crop in crops])    # 5, B, C, H, W
        x = torch.cat([crops, flips])                                       # 10, B, C, H, W
        x = rearrange(x, 'n b c h w -> (n b) c h w')                        # (10*B), C, H, W
        y = self.forward(x)                                                 # (10*B), N
        y = rearrange(y, '(n b) c -> n b c', b=B)                           # 10, B, N
        return y.mean(dim=0)                                                # B, N
    # paper repots top-1 and top-5 test error rate of 40.7% and 18.2% en imagenet 2012 validation set


from torch.utils.data import Dataset
from skimage import io 
import numpy as np
from einops import rearrange

class ImageClassificationDataset(Dataset):
    def __init__(self, images, labels=None, trans=None):
        self.images = images 
        self.labels = labels 
        self.trans = trans 

    def __len__(self):
        return len(self.images) 

    def __getitem__(self, idx):
        im = io.imread(self.images[idx]).astype(np.float32) / 255.
        im = self.apply_transforms(im)
        im = rearrange(im, 'h w c -> c h w')
        if self.labels is not None:
            return im, self.labels[idx]
        return im

    # by default, we assume albumentations
    # to use another implementation, override this method
    def apply_transforms(self, im):
        if self.trans is not None:
            return self.trans(image=im)['image']
        return im


import os 
from glob import glob 
from pathlib import Path
import pandas as pd
import xml.etree.ElementTree as ET
from tqdm import tqdm
import multiprocessing
from concurrent.futures import ThreadPoolExecutor
from PIL import Image
import argparse
from torch.utils.data import Dataset
import pandas as pd
from skimage import io
import numpy as np
from einops import rearrange

# resize image by keeping the aspect ratio and cropping the center
def resize_and_crop(path, size=256):
    img = Image.open(path).convert('RGB') # some images are RGBA
    w, h = img.size
    ar = w / h
    if ar > 1:
        img = img.resize((int(size * ar), size), Image.BILINEAR)
        w, h = img.size
        return img.crop((w//2 - size//2, 0, w//2 + size//2, size))
    img = img.resize((size, int(size / ar)), Image.BILINEAR)
    w, h = img.size
    return img.crop((0, h//2 - size//2, size, h//2 + size//2))	

# preprocess a single image and save the result
def preprocess(args):
    dst_path, mode, path, size, cls = args
    img = resize_and_crop(path, size)
    img_name = path.split('/')[-1]
    if cls is None:
        dst_folder = f'{dst_path}/{mode}'
    else:
        dst_folder = f'{dst_path}/{mode}/{cls}'
    os.makedirs(dst_folder, exist_ok=True)
    new_path = f'{dst_folder}/{img_name}'
    img.save(new_path)
    return new_path

# this function will process the entire dataset, resizing and croping the images and generating the output folders for training
# the datasets is supposed to be downloaded from https://www.kaggle.com/competitions/imagenet-object-localization-challenge/data

def process(base_path = 'data/ILSVRC', dst_path = 'data/imagenet256', size=256, workers=None):
    # this is where the dataset is supposed to be
    path = f'{base_path}/Data/CLS-LOC/'
    classes = sorted(os.listdir(f'{path}/train'))
    assert len(classes) == 1000
    # create df with all the training images and their labels
    print("Generating training set...")
    images, labels, cls = [], [], []
    for i, c in enumerate(tqdm(classes)):
        this_images = glob(f'{path}/train/{c}/*.JPEG')
        images += this_images
        labels += [i] * len(this_images)
        cls += [c] * len(this_images)
    train = pd.DataFrame({'image': images, 'label': labels, 'class': cls})
    assert len(train) == 1281167
    # repeat for validation
    print("Generating validation set...")
    val_annotations = glob(f'{base_path}/Annotations/CLS-LOC/val/*.xml')
    val_images, val_labels, val_cls = [], [], []
    for ann in tqdm(val_annotations):
        # read xml
        tree = ET.parse(ann)
        root = tree.getroot()
        cls = root.findall('object')[0].find('name').text
        val_cls.append(cls)
        val_labels.append(classes.index(cls))
        val_images.append(f'{path}/val/{ann.split("/")[-1].split(".")[0]}.JPEG')
    val = pd.DataFrame({'image': val_images, 'label': val_labels, 'class': val_cls})
    assert len(val.label.unique()) == len(classes)
    assert len(val) == 50000
    # repeat for test
    test = glob(f'{path}/test/*.JPEG')
    assert len(test) == 100000
    # process images in parallel
    args = [(dst_path, 'train', img, size, cls) for img, cls in zip(train.image.values, train['class'].values)] + \
        [(dst_path, 'val', img, size, cls) for img, cls in zip(val.image.values, val['class'].values)] + \
        [(dst_path, 'test', img, size, None) for img in test]
    print("Processing images...")
    num_cores = multiprocessing.cpu_count() if workers is None else workers
    with ThreadPoolExecutor(max_workers=num_cores) as pool:
        with tqdm(total=len(args)) as progress:
            futures = []
            for arg in args:
                future = pool.submit(preprocess, arg) # enviamos la tupla de argumentos
                future.add_done_callback(lambda p: progress.update())
                futures.append(future)
            results = []
            for future in futures:
                result = future.result()
                results.append(result)
    assert len(results) == len(train) + len(val) + len(test)

def compute_mean_std(image):
    image = io.imread(image).astype(np.float32) / 255
    return image.mean(axis=(0,1)), image.std(axis=(0,1))

def compute_stats(images, workers=None):
    num_cores = multiprocessing.cpu_count() if workers is None else workers
    with ThreadPoolExecutor(max_workers=num_cores) as pool:
        with tqdm(total=len(images)) as progress:
            futures = []
            for img in images:
                future = pool.submit(compute_mean_std, img) 
                future.add_done_callback(lambda p: progress.update())
                futures.append(future)
            results = []
            for future in futures:
                result = future.result()
                results.append(result)
    mean, std = zip(*results)
    return np.array(mean).mean(axis=0), np.array(std).mean(axis=0)

class ImageNet(Dataset):
    def __init__(self, path, mode, trans=None, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]):
        assert mode in ['train', 'val', 'test'], 'mode must be train, val or test'
        path = Path(path)
        self.labels = None
        if mode == 'test':
            self.images = glob(str(path/'test/*.JPEG'))
            assert len(self.images) == 100000
        else:
            classes = sorted(os.listdir(path/mode))
            assert len(classes) == 1000
            images, labels = [], []
            for i, c in enumerate(classes):
                this_images = glob(str(path/mode/c/'*.JPEG'))
                images += this_images
                labels += [i] * len(this_images)
            if mode == 'train':
                assert len(images) == 1281167
            else:
                assert len(images) == 50000
            self.images, self.labels = images, labels
        self.trans = trans 
        self.mean = np.array(mean).astype(np.float32)
        self.std = np.array(std).astype(np.float32)

    def __len__(self):
        return len(self.images)
    
    def __getitem__(self, ix):
        # read image
        im = io.imread(self.images[ix])
        # apply transformations on uint8 RGB images
        if self.trans is not None:
            im = self.trans(image=im)['image']
        # normalize
        im = im.astype(np.float32) / 255.
        im = (im - self.mean) / self.std
        im = rearrange(im, 'h w c -> c h w')
        # read label if available
        if self.labels is not None:
            return im, self.labels[ix]
        return im
        

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process imagenet.')
    parser.add_argument('--base-path', help='Folder with the original imagenet dataset')
    parser.add_argument('--dst-path', help='Folder where the processed dataset will be saved')
    parser.add_argument('--size', help='Size of the images', default=256)
    parser.add_argument('--workers', help='Number of workers to use', default=None)
    args = parser.parse_args()
    process(
        base_path=args.base_path,
        dst_path=args.dst_path,
        size=args.size,
        workers=args.workers
    )


import requests
import os
import pandas as pd
import zipfile
from tqdm import tqdm

def download_url(url, save_path, chunk_size=128):
    r = requests.get(url, stream=True)
    with open(save_path, 'wb') as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)

def unzip_file(source, path, msg="Extracting"):
    with zipfile.ZipFile(source) as zf:
        for member in tqdm(zf.infolist(), desc=msg):
            try:
                zf.extract(member, path)
            except zipfile.error as e:
                pass

def download_data(path, compressed_data_filename, data_folder, url, verbose):
    compressed_data_path = path / compressed_data_filename
    uncompressed_data_path = path / data_folder
    # create data folder
    os.makedirs(path, exist_ok=True)
    # extract
    if not os.path.isdir(uncompressed_data_path):
        # check data is not already downloaded
        if not os.path.isfile(compressed_data_path):
            print("downloading data ...")
            download_url(url, compressed_data_path)
        else:
            print("data already downloaded !")
        unzip_file(compressed_data_path, path, msg="extracting data ...")
    else:
        if verbose:
            print("data already downloaded and extracted !")
    return uncompressed_data_path


# retrieve classes from folder structure	
def generate_classes_list(uncompressed_data_path):
    return sorted(os.listdir(uncompressed_data_path))


def generate_df(classes, uncompressed_data_path, verbose):
    images, labels = [], []
    for ix, label in enumerate(classes):
        _images = os.listdir(uncompressed_data_path / label)
        images += [str(uncompressed_data_path /
                       label / img) for img in _images]
        labels += [ix]*len(_images)
    assert len(images) == len(labels)
    if verbose:
        print(f'Number of images: {len(images)}')
    return pd.DataFrame({'image': images, 'label': labels})




import torch
from fastprogress.fastprogress import master_bar, progress_bar
import numpy as np
from ..metrics import accuracy

def fit(
    model, 
    dataloader, 
    optimizer, 
    criterion, 
    metrics={'acc': accuracy},
    max_epochs=20,
    # debug
    overfit_batches=0,
    limit_train_batches=0,
    limit_val_batches=0,
    after_epoch_log=True,
    # callbacks
    on_epoch_end=lambda h,m,o,e: None,
    after_val=lambda vl: None,
    # device
    device="cpu", 
    rank=0,
    compile=False,
    *args,
    **kwargs
):
    if device == "cuda":
        device_type = "cuda"
        device = f"cuda:{rank}"
    else: 
        device_type = "cpu"
    print("Training model on", device)
    # count parameters
    if rank == 0:
        n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
        print(f"Number of trainable parameters: {n_params}")
    if compile:
        torch.backends.cuda.matmul.allow_tf32 = True # allow tf32 on matmul
        torch.backends.cudnn.allow_tf32 = True # allow tf32 on cudnn
        # torch.set_float32_matmul_precision('high')
        print(rank, "Compiling model ...") # it can takw a while...
        model = torch.compile(model)
    model.to(device)
    mb = master_bar(range(1, max_epochs+1)) if rank == 0 else range(1, max_epochs+1)
    hist = {'epoch': [], 'loss': [], 'lr': []}
    for metric in metrics.keys():
        hist[metric] = []
    if not overfit_batches and 'val' in dataloader:
        hist['val_loss'] = []
        for metric in metrics.keys():
            hist['val_' + metric] = []
    for epoch in mb:
        model.train()
        train_logs = {'loss': []}
        for metric in metrics.keys():
            train_logs[metric] = []
        hist['epoch'].append(epoch)
        pbar = progress_bar(dataloader['train'], parent=mb) if rank == 0 else dataloader['train']
        for batch_ix, batch in enumerate(pbar):
            X, y = batch
            X, y = X.to(device), y.to(device)
            optimizer.zero_grad()
            with torch.autocast(device_type=device_type, dtype=torch.bfloat16):
                y_hat = model(X)
                loss = criterion(y_hat, y)
            loss.backward()
            optimizer.step()
            train_logs['loss'].append(loss.item())
            for metric in metrics.keys():
                train_logs[metric].append(metrics[metric](y_hat, y).item())
            log = f"loss {np.mean(train_logs['loss']):.5f}"
            for metric in metrics.keys():
                log += f" {metric} {np.mean(train_logs[metric]):.5f}"
            if rank == 0: mb.child.comment = log
            if overfit_batches and batch_ix > overfit_batches:
                break
            if limit_train_batches and batch_ix > limit_train_batches:
                break
        hist['loss'].append(np.mean(train_logs['loss']))
        for metric in metrics.keys():
            hist[metric].append(np.mean(train_logs[metric]))
        if not overfit_batches and 'val' in dataloader:
            val_logs = {'loss': []}
            for metric in metrics.keys():
                val_logs[metric] = []
            model.eval()
            with torch.no_grad():
                pbar = progress_bar(dataloader['val'], parent=mb) if rank == 0 else dataloader['val']
                for batch_ix, batch in enumerate(pbar):
                    X, y = batch
                    X, y = X.to(device), y.to(device)
                    with torch.autocast(device_type=device_type, dtype=torch.bfloat16):
                        y_hat = model(X)
                        loss = criterion(y_hat, y)
                    val_logs['loss'] .append(loss.item())
                    for metric in metrics.keys():
                        val_logs[metric].append(metrics[metric](y_hat, y).item())
                    if rank == 0:
                        _log = f"val_loss {np.mean(val_logs['loss']):.5f}"
                        for metric in metrics.keys():
                            _log += f" val_{metric} {np.mean(val_logs[metric]):.5f}"
                        mb.child.comment = _log
                    if limit_val_batches and batch_ix > limit_val_batches:
                        break
            hist['val_loss'].append(np.mean(val_logs['loss']))
            for metric in metrics.keys():
                hist['val_' + metric].append(np.mean(val_logs[metric]))
            if rank == 0:
                log += f" val_loss {np.mean(val_logs['loss']):.5f}"
                for metric in metrics.keys():
                    log += f" val_{metric} {np.mean(val_logs[metric]):.5f}"
            after_val(val_logs)
        hist['lr'].append(optimizer.param_groups[0]['lr'])
        if rank == 0: mb.main_bar.comment = log
        if after_epoch_log and rank == 0: 
            mb.write(f"Epoch {epoch}/{max_epochs} " + log)
        on_epoch_end(hist, model, optimizer)
    return hist


from torch.utils.data import Dataset
from skimage import io 
import numpy as np
from einops import rearrange

class SSLDataset(Dataset):
    def __init__(self, images, bands, trans=None):
        self.images = images 
        self.trans = trans 
        self.bands = bands

    def __len__(self):
        return len(self.images) 

    def __getitem__(self, idx):
        bands = []
        for band in self.bands:
            bands.append(io.imread(f'{self.images[idx]}/{band}.tif'))
        im = np.stack(bands, axis=-1).astype(np.float32) / 255.
        im1 = self.apply_transforms(im)
        im1 = rearrange(im1, 'h w c -> c h w')
        im2 = self.apply_transforms(im)
        im2 = rearrange(im2, 'h w c -> c h w')
        return im1, im2

    # by default, we assume albumentations
    # to use another implementation, override this method
    def apply_transforms(self, im):
        if self.trans is not None:
            return self.trans(image=im)['image']
        return im


from pathlib import Path
import pandas as pd
import os
from glob import glob

# https://github.com/ServiceNow/seasonal-contrast

class SeCo100k():
    def __init__(self, path="/fastdata/seasonal_contrast_100k"):
        super().__init__()
        self.path = Path(path)
        self.in_chans = 3
        images = glob(str(self.path) + '/*/*.tif')
        self.data = pd.DataFrame({'image': images})


class SeCo():
    def __init__(self, path="/fastdata/seasonal_contrast_1m", bands=['B4', 'B3', 'B2']):
        super().__init__()
        self.path = Path(path)
        self.in_chans = len(bands)
        self.bands = bands
        aois = os.listdir(str(self.path))
        images = []
        for aoi in aois:
            images += [f'{self.path}/{aoi}/{f}' for f in os.listdir(str(self.path / aoi))]
        self.data = pd.DataFrame({'image': images})


import random 
import torch 
import numpy as np

def seed_everything(seed=42):
	random.seed(seed)
	np.random.seed(seed)
	torch.manual_seed(seed)
	torch.cuda.manual_seed(seed)
	torch.backends.cudnn.deterministic = True
	torch.backends.cudnn.benchmark = False
    


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['convnets',
 'convnets.datasets',
 'convnets.models',
 'convnets.train',
 'convnets.train.imagenet']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'convnets',
    'version': '0.2.5',
    'description': 'Convolutional Neural Networks and utilities for Computer Vision',
    'long_description': '# convnets\n\nConvolutional Neural Networks and utilities for Computer Vision.\n\n🚧 Under construction',
    'author': 'juansensio',
    'author_email': 'sensio.juan@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import torch
from fastprogress.fastprogress import master_bar, progress_bar
import numpy as np

def barlow_step(model, batch, device, l=5e-3):
    # two randomly augmented versions of x
    x1, x2 = batch
    x1, x2 = x1.to(device), x2.to(device)

    # compute representations
    z1 = model(x1)
    z2 = model(x2)

    # normalize repr. along the batch dimension
    N, D = z1.shape
    z1_norm = (z1 - z1.mean(0)) / z1.std(0) # NxD
    z2_norm = (z2 - z2.mean(0)) / z2.std(0) # NxD

    # cross-correlation matrix
    c = (z1_norm.T @ z2_norm) / N # DxD

    # loss
    c_diff = (c - torch.eye(D, device=device)).pow(2) # DxD
    # multiply off-diagonal elems of c_diff by lambda
    d = torch.eye(D, dtype=bool)
    c_diff[~d] *= l
    return c_diff.sum()

def barlow_fit(model, dataloader, optimizer, scheduler=None, use_amp = True, epochs=10, device="cuda", log=True, eval_each=10, limit_train_batches=0, ssl_eval=None):
    print("Training model on", device)
    model.to(device)
    mb = master_bar(range(1, epochs+1))
    hist = {'loss': [], 'epoch': []}
    scaler = torch.cuda.amp.GradScaler(enabled=use_amp)
    for epoch in mb:
        # train
        model.train()
        train_loss = []
        hist['epoch'].append(epoch)
        for batch_ix, batch in enumerate(progress_bar(dataloader, parent=mb)):
            optimizer.zero_grad()
            with torch.cuda.amp.autocast(enabled=use_amp):
                loss = barlow_step(model, batch, device)
            scaler.scale(loss).backward()
            # gradient clipping
            #torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=0.1)
            scaler.step(optimizer)
            scaler.update()
            train_loss.append(loss.item())
            mb.child.comment = f"loss {np.mean(train_loss):.5f}"
            if limit_train_batches and batch_ix > limit_train_batches:
                break
        hist['loss'].append(np.mean(train_loss))
        _log = f"loss {np.mean(train_loss):.5f}"
        scheduler.step()
        # eval
        if ssl_eval is not None and not epoch % eval_each:
            print("evaluating ...")
            ssl_eval(model)
        mb.main_bar.comment = _log
        if log: 
            mb.write(f"Epoch {epoch}/{epochs} " + _log)
        torch.save(model.state_dict(), f"./checkpoints/barlow.pth")
    return hist